# Copyright (C) 2008 Ansgar Wiechers <ansgar.wiechers@planetcobalt.net>
#
# This module is free software. You may use, redistribute and/or modify it
# under the terms of the GNU General Public License Version 2.0 as found at
# <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>.

package PostgreSQL;
use FilterDB::DB;
our @ISA = qw( DB );

=head1 NAME

  FilterDB::PostgreSQL -- Backscatter filter backend (PostgreSQL)

=head1 SYNOPSIS

  use FilterDB::PostgreSQL;
  use strict;

  my $db = new PostgreSQL(
      'db_pass'    => 'baz',
      'read_write' => 1,
    );  

  if ($db->check('foo')) {
    print "'foo' found\n";
  } else {
    print "'foo' not found\n";
  }

  $db->insert('foo');
  $db->insert('bar');

  if ($db->check('foo')) {
    print "'foo' found\n";
  } else {
    print "'foo' not found\n";
  }

  my %records = $db->list();
  my @keys = keys %records;
  if ($#keys >= 0) {
    print "$records{$key}: $key\n" foreach my $key (@keys);
  }

=head1 REQUIRES

Perl5.8.8, DBD::Pg

=head1 DESCRIPTION

The PostgreSQL class implements a PostgreSQL backend for the backscatter filter.
It stores MD5 hashes of given values and the timestamps when the values were
entered into the database. It provides methods to check whether the hash for any
given value is present in the database, to add insert new values, to list the
database contents, and to remove outdated values.

Status information will be logged to syslog if the calling application has
opened the syslog via openlog().

=head1 METHODS

=head2 Creation

=over 4

=item new($db_pass, [$db_name], [$db_host], [$db_port], [$db_user], [$client_encoding], [$read_write], [$max_age])

Opens a connection to a PostgreSQL database with the given parameters:

  $db_pass (required)          Password for accessing the database
  $db_user (optional)          Username for accessing the database
                               (default: name of the user in whose
                               context the process is running)
  $db_host (optional)          Database host (default: 127.0.0.1)
  $db_port (optional)          Database port (default: 5432)
  $db_name (optional)          Name of the database
                               (default: filterdb)
  $sslmode (optional)          Use SSL-encrypted connections?
                               (default: prefer)
  $client_encoding (optional)  Encoding the client is using
                               (default: LATIN1)
  $read_write (optional)       Open the database for reading if FALSE
                               or writing if TRUE (default: FALSE)
  $max_age (optional)          Number of days after which entries
                               should be removed from the database
                               by cleanup() (default: 4)

=back

=head2 Access

=over 4

=item check($value)

Checks whether the MD5 hash of the given $value is present in the database.
Returns TRUE if the hash is present, otherwise FALSE.

=item cleanup()

Removes entries older than $max_age days from the database.

=item insert($value)

Inserts the MD5 hash of the given $value and the current timestamp into the
database. If the hash is already present, the existing entry is updated with the
durrent timestamp.

=item list()

Returns a dictionary with all MD5 hash->timestamp pairs stored in the database.

=back

=head1 AUTHOR

Ansgar Wiechers <ansgar.wiechers@planetcobalt.net>

=cut

require 5.8.8;

use strict;
use warnings;
use DBI;
use Sys::Syslog qw( :standard :macros );

use constant TRUE	=> 1;
use constant FALSE => not TRUE;

my $driver = 'dbi:Pg';
my $default_host = '127.0.0.1';
my $default_port = '5432';
my $default_name = 'filterdb';
my $default_sslmode = 'prefer';
my $default_encoding = 'LATIN1';
my $default_max_age = 4; # days

my $check_statement   = "SELECT hash FROM hashes WHERE hash = md5(?)";
my $insert_statement  = "INSERT INTO hashes (hash, created) VALUES (md5(?), date_trunc('second', now()))";
my $update_statement  = "UPDATE hashes SET created = date_trunc('second', now()) WHERE hash = md5(?)";
my $cleanup_statement = "DELETE FROM hashes WHERE now() - created > CAST(? AS integer)";
my $list_statement    = "SELECT hash, created FROM hashes";

sub new {
	my ($class, %args) = @_;
	my $self = bless { %args }, $class;

	$self->{db_host} = $default_host unless defined $self->{db_host};
	$self->{db_port} = $default_port unless defined $self->{db_port};
	$self->{db_name} = $default_name unless defined $self->{db_name};
	$self->{db_user} = $ENV{'USER'} unless defined $self->{db_user};
	$self->{db_pass} = '' unless defined $self->{db_pass};
	$self->{sslmode} = $default_sslmode unless defined $self->{sslmode};
	$self->{client_encoding} = $default_encoding unless defined $self->{client_encoding};
	$self->{read_write} = ($self->{read_write}) ? TRUE : FALSE;
	$self->{max_age} = $default_max_age unless defined $self->{max_age} and $self->{max_age} =~ m/^[0-9]+$/;

	my $transaction_mode = ($self->{read_write}) ? 'READ WRITE' : 'READ ONLY';

	syslog(LOG_DEBUG, "Opening database connection [$self->{db_user}\@$self->{db_host}:$self->{db_port}/$self->{db_name}] for " . (($self->{read_write}) ? 'writing' : 'reading'));

	$self->{db} = DBI->connect("$driver:dbname=$self->{db_name};host=$self->{db_host};port=$self->{db_port};sslmode=$self->{sslmode};", $self->{db_user}, $self->{db_pass}, {AutoCommit => FALSE})
		or _terminate('Connect to database failed: ' . DBI->errstr . "\n");
	$self->{db}->do("SET TRANSACTION $transaction_mode;")
		or _terminate('Failed to set transaction mode to "' . lc($transaction_mode) . '": ' . DBI->errstr . "\n");
	$self->{db}->do("SET client_encoding = '$self->{client_encoding}';")
		or _terminate('Failed to set client encoding to "' . lc($self->{client_encoding}) . '": ' . DBI->errstr . "\n");

	$self->{check_query} = $self->{db}->prepare($check_statement)
		or _terminate("Preparing SELECT statement failed: " . DBI->errstr . "\n");
	$self->{insert_query} = $self->{db}->prepare($insert_statement)
		or _terminate("Preparing INSERT INTO statement failed: " . DBI->errstr . "\n");
	$self->{update_query} = $self->{db}->prepare($update_statement)
		or _terminate("Preparing UPDATE statement failed: " . DBI->errstr . "\n");
	$self->{cleanup_query} = $self->{db}->prepare($cleanup_statement)
		or _terminate("Preparing DELETE statement failed: " . DBI->errstr . "\n");
	$self->{list_query} = $self->{db}->prepare($list_statement)
		or _terminate("Preparing SELECT * statement failed: " . DBI->errstr . "\n");

	return $self;
}

sub DESTROY {
	my ($self) = @_;
	syslog(LOG_DEBUG, "Closing database connection");
	if ($self->{db}) {
		$self->{db}->rollback;
		$self->{db}->disconnect;
	}
}

sub check {
	my ($self, $value) = @_;
	my $hash;

	_terminate('check(): Missing argument $value') unless defined $value;

	syslog(LOG_DEBUG, "Checking for presence of '$value'");

	$self->{check_query}->execute($value)
		or _terminate("SELECT failed: " . DBI->errstr . "\n");
	$self->{check_query}->bind_columns(undef, \$hash)
		or _terminate("Binding result column failed: " . DBI->errstr . "\n");
	$self->{check_query}->fetchrow_arrayref();
	$self->{check_query}->finish;
	$self->{db}->rollback;

	return ((defined $hash) ? TRUE : FALSE);
}

sub cleanup {
	my ($self) = @_;

	syslog(LOG_DEBUG, "Removing entries older than $self->{max_age} days");

	$self->{cleanup_query}->execute($self->{max_age})
		or _terminate("DELETE failed: " . DBI->errstr . "\n");
	$self->{cleanup_query}->finish;
	$self->{db}->commit;
}

sub insert {
	my ($self, $value) = @_;
	my $query;

	_terminate('insert(): Missing argument $value') unless defined $value;

	if ($self->check($value)) {
		syslog(LOG_DEBUG, "Updating record for '$value' in database");
		$query = $self->{update_query};
	} else {
		syslog(LOG_DEBUG, "Inserting '$value' into database");
		$query = $self->{insert_query};
	}
	$query->execute($value)
		or _terminate("INSERT/UPDATE failed: " . DBI->errstr . "\n");
	$query->finish;
	$self->{db}->commit;
}

sub list {
	my ($self) = @_;
	my %records;
	my ($hash, $timestamp);

	syslog(LOG_DEBUG, 'Listing database entries');

	$self->{list_query}->execute()
		or _terminate("SELECT failed: " . DBI->errstr . "\n");
	$self->{list_query}->bind_columns(undef, \$hash, \$timestamp)
		or _terminate("Binding result columns failed: " . DBI->errstr . "\n");
	while ($self->{list_query}->fetchrow_arrayref()) {
		$records{$hash} = $timestamp;
	}
	$self->{list_query}->finish;
	$self->{db}->rollback;

	return %records;
}

sub _terminate($) {
	syslog(LOG_ERR, $_[0]);
	die $_[0];
}

1;
